/*  HSYCO.cpp - HSYCO Library for Arduino
	Copyright (C) 2012-2013 Home Systems Consulting S.p.A. - All rights reserved.

	For information, see the Home Systems Consulting web site:
	http://www.homesystemsconsulting.com/
	
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.
    See file LICENSE.txt for further informations on licensing terms.
*/

#include "HSYCO.h"
#include "HardwareSerial.h"

extern "C" {
#include <string.h>
#include <stdlib.h>
}

/* constructor */
HSYCOClass::HSYCOClass(Stream &s) : HSYCOSerial(s) {
	currentCallback = NULL;
	init();
}

/* public methods */
void HSYCOClass::begin(void) {
	begin(115200);
}

void HSYCOClass::begin(long speed) {
	Serial.begin(speed);
	HSYCOSerial = Serial;
}

void HSYCOClass::begin(Stream &s) {
	HSYCOSerial = s;
	inputBufferReset();
}

void HSYCOClass::debug(void) {
	debugMode = true;
}

void HSYCOClass::process(void) {
	while (HSYCOSerial.available()) {
		unsigned char inChar = (char)HSYCOSerial.read();
		if (debugMode || inputStateStart) {
			if (inChar == '\n') {
				inputStateStart = false;
				inputBufferCommit();
				if (inputValueBufferLength == 0 && strcmp(inputNameBuffer, "ping") == 0) {
					if (!debugMode) {
						HSYCOSerial.write(255);
					}
					HSYCOSerial.write("ping\n");
				} else if (inputValueBufferLength == 0 && strcmp(inputNameBuffer, "version") == 0) {
					if (!debugMode) {
						HSYCOSerial.write(255);
					}
					HSYCOSerial.write("V.");
					HSYCOSerial.write(HSYCO_VERSION);
					HSYCOSerial.write('\n');
				} else if (strcmp(inputNameBuffer, "pinmode") == 0) {
					String params = String(inputValueBuffer);
					int sep = params.indexOf(',');
					if (sep > 0) {
						inputValueBuffer[sep] = '\0';
						char *mode = inputValueBuffer + sep + 1;
						char *p;
						int pin = (inputValueBuffer[0] == 'a') ? A0 + (int)strtol(inputValueBuffer + 1, &p, 10) : (int)strtol(inputValueBuffer, &p, 10);
						if (!*p) {
							if (strcmp(mode, "input") == 0) {
								pinMode(pin, INPUT);
							} else if (strcmp(mode, "output") == 0) {
								pinMode(pin, OUTPUT);
							} else if (strcmp(mode, "input_pullup") == 0) {
								pinMode(pin, INPUT_PULLUP);
							}
						}
					}
				} else if (strcmp(inputNameBuffer, "digitalwrite") == 0) {
					String params = String(inputValueBuffer);
					int sep = params.indexOf(',');
					if (sep > 0) {
						inputValueBuffer[sep] = '\0';
						char *mode = inputValueBuffer + sep + 1;
						char *p;
						int pin = (inputValueBuffer[0] == 'a') ? A0 + (int)strtol(inputValueBuffer + 1, &p, 10) : (int)strtol(inputValueBuffer, &p, 10);
						if (!*p) {
							if (strcmp(mode, "1") == 0) {
								digitalWrite(pin, HIGH);
								if (!debugMode) {
									HSYCOSerial.write(255);
								}
								HSYCOSerial.print("pin.");
								HSYCOSerial.print(inputValueBuffer);
								HSYCOSerial.print("=1\n");
							} else if (strcmp(mode, "0") == 0) {
								digitalWrite(pin, LOW);
								if (!debugMode) {
									HSYCOSerial.write(255);
								}
								HSYCOSerial.print("pin.");
								HSYCOSerial.print(inputValueBuffer);
								HSYCOSerial.print("=0\n");
							}
						}
					}
				} else if (strcmp(inputNameBuffer, "digitalread") == 0) {
					char *p;
					int pin = (inputValueBuffer[0] == 'a') ? A0 + (int)strtol(inputValueBuffer + 1, &p, 10) : (int)strtol(inputValueBuffer, &p, 10);
					if (!*p) {
						if (!debugMode) {
							HSYCOSerial.write(255);
						}
						HSYCOSerial.print("pin.");
						HSYCOSerial.print(inputValueBuffer);
						if (digitalRead(pin) == HIGH) {
							HSYCOSerial.print("=1\n");
						} else {
							HSYCOSerial.print("=0\n");
						}
					}
				} else if (strcmp(inputNameBuffer, "analogwrite") == 0) {
					String params = String(inputValueBuffer);
					int sep = params.indexOf(',');
					if (sep > 0) {
						int value = atoi(inputValueBuffer + sep + 1);
						inputValueBuffer[sep] = '\0';
						char *p;
						int pin;
						if (strncmp(inputValueBuffer, "dac", 3) == 0) {
							if (HSYC0_DAC0 != -1) {
								pin =  HSYC0_DAC0 + (int)strtol(inputValueBuffer + 3, &p, 10);
							}
						} else if (inputValueBuffer[0] == 'a') {
							pin =  A0 + (int)strtol(inputValueBuffer + 1, &p, 10);
						} else {
							pin =  (int)strtol(inputValueBuffer, &p, 10);
						}
						if (!*p) {
							if (!debugMode) {
								HSYCOSerial.write(255);
							}
							analogWrite(pin, value);
							HSYCOSerial.print("pin.");
							HSYCOSerial.print(inputValueBuffer);
							HSYCOSerial.print('=');
							HSYCOSerial.print(value);
							HSYCOSerial.print('\n');
						}
					}
				} else if (strcmp(inputNameBuffer, "analogread") == 0) {
					char *p;
					int pin = (inputValueBuffer[0] == 'a') ? A0 + (int)strtol(inputValueBuffer + 1, &p, 10) : (int)strtol(inputValueBuffer, &p, 10);
					if (!*p) {
						if (!debugMode) {
							HSYCOSerial.write(255);
						}
						HSYCOSerial.print("pin.");
						HSYCOSerial.print(inputValueBuffer);
						HSYCOSerial.print('=');
						HSYCOSerial.print(analogRead(pin));
						HSYCOSerial.print('\n');
					}
				} else if (inputStateName && currentCallback != NULL && strncmp(inputNameBuffer, "call:", 5) == 0) {
					(*currentCallback)(inputNameBuffer + 5, inputValueBuffer);
				}
				inputBufferReset();
			} else if (inChar == '=') {
				inputBufferCommit();
				inputStateName = true;
			} else {
				inputBufferAppend(inChar);
			}
		} else if (inChar == 255) {
			inputStateStart = true;
		}
	}
}

void HSYCOClass::ioSet(char* name, char* value) {
	if (!debugMode) {
		HSYCOSerial.write(255);
	}
	HSYCOSerial.write(name);
	HSYCOSerial.write('=');
	HSYCOSerial.write(value);
	HSYCOSerial.write('\n');
}

void HSYCOClass::ioSet(char* name, int value) {
	if (!debugMode) {
		HSYCOSerial.write(255);
	}
	HSYCOSerial.write(name);
	HSYCOSerial.write('=');
	HSYCOSerial.print(value);
	HSYCOSerial.write('\n');
}

void HSYCOClass::attach(callbackFunction myFunction) {
	currentCallback = myFunction;
}

void HSYCOClass::detach(void) {
	currentCallback = NULL;
}

/* private methods */
void HSYCOClass::init(void) {
	inputBufferInit();
}

void HSYCOClass::inputBufferInit(void) {
	inputStateName = false;
	inputNameBufferSize = 33;
	inputValueBufferSize = 33;
	inputNameBufferLength = 0;
	inputValueBufferLength = 0;
	inputNameBuffer = (char*)malloc(inputNameBufferSize);
	inputValueBuffer = (char*)malloc(inputValueBufferSize);
}

void HSYCOClass::inputBufferReset(void) {
	inputStateName = false;
	inputNameBufferLength = 0;
	inputValueBufferLength = 0;
}

void HSYCOClass::inputBufferAppend(char c) {
	if (!inputStateName) {
		if (inputNameBufferLength + 1 >= inputNameBufferSize) {
			inputNameBufferSize += 32;
			inputNameBuffer = (char*)realloc(inputNameBuffer, inputNameBufferSize);
		}
		inputNameBuffer[inputNameBufferLength] = c;
		inputNameBufferLength++;
	} else {
		if (inputValueBufferLength + 1 >= inputValueBufferSize) {
			inputValueBufferSize += 32;
			inputValueBuffer = (char*)realloc(inputValueBuffer, inputValueBufferSize);
		}
		inputValueBuffer[inputValueBufferLength] = c;
		inputValueBufferLength++;
	}
}

void HSYCOClass::inputBufferCommit(void) {
	if (!inputStateName) {
		inputNameBuffer[inputNameBufferLength] = '\0';
		inputValueBuffer[0] = '\0';
	} else {
		inputValueBuffer[inputValueBufferLength] = '\0';
	}
}

// create the instance
HSYCOClass HSYCO(Serial);
